#include <Arduino.h>
#include <DFRobot_ID809.h>
#include <EEPROM.h>
#include <sha256.h>

// use serial3 as interface with fingerprint
// sensor.

#define FPSerial Serial3

// define buillt_in led

#define BUILTIN_LED 13

// Collect the fingerprint 3 times

#define COLLECT_NUMBER 3
#define SHA256_SIZE 32
// EEPROM OFFSET to store

/*------Global variables--------------*/

DFRobot_ID809 fingerprint;
uint8_t ID = 0, ret;
byte hash[SHA256_SIZE];
uint8_t temp[1008];
const uint16_t EEPROM_FP_HASH = 1696;

/*---------Compute hash----------------*/

void compute_hash() {

  SHA256_CTX ctx;
  sha256_init(&ctx);
  sha256_update(&ctx, temp, sizeof(temp));
  sha256_final(&ctx, hash);

  for (uint8_t i = 0; i < SHA256_SIZE; i++) {

    if (hash[i] < 0x10) {

      Serial.print('0');
    }

    Serial.print(hash[i], HEX);
    Serial.print(" ");
  }
  Serial.println("\n");
  Serial.println("-----------------");
}

/*-----Blink on board led -------------- */

void blink_led() {
  digitalWrite(BUILTIN_LED, HIGH);
  delay(500);
  digitalWrite(BUILTIN_LED, LOW);
  delay(500);
}

/*-------- Save FingerPrint hash in EEPROM ------*/

void saveHash() {
  uint16_t memoryAddress;

  memoryAddress = EEPROM_FP_HASH + ((ID - 1) * 32);

  if(memoryAddress > 2016){
  	Serial.println("Error: not enough space in" 
			" EEPROM");
	return;
  }

  for (uint8_t i = 0; i < SHA256_SIZE; i++) {

    EEPROM.write(memoryAddress, hash[i]);
    memoryAddress++;
  }
}

/*---------Arduino setup function------*/

void setup() {
  // config led_pin as output
  pinMode(BUILTIN_LED, OUTPUT);
  // use serial port to print the baud rate
  Serial.begin(9600);
  // config fingerprint sensor
  FPSerial.begin(115200);
  fingerprint.begin(FPSerial);

  while (fingerprint.isConnected() == false) {
    Serial.println("Communication with the device failed");
    delay(1000);
  }
  // Disable self-learn feature
  fingerprint.setSelfLearn(0);
}

/*------Read stored EEPROM data--------- */

void readHash() {

  uint16_t memoryAddress;
  byte eeprom_data;

  memoryAddress = EEPROM_FP_HASH + ((ID - 1) * 32);

  for (uint16_t i = memoryAddress; i < (memoryAddress + SHA256_SIZE); i++) {

    eeprom_data = EEPROM.read(i);
    Serial.print(eeprom_data, HEX);
    Serial.print(" ");
  }
  Serial.println("\n");
  Serial.println("--------------------");
}

/*---------Register Finger print ------------*/

void regFingerprint() {
  uint8_t i = 0;
  /*Get an unregistered ID for saving fingerprint
Return ID when succeeded
Return ERR_ID809 if failed
*/
  if ((ID = fingerprint.getEmptyID()) == ERR_ID809) {
    while (1) {
      /*Get error code information*/
      // desc = fingerprint.getErrorDescription();
      // Serial.println(desc);
      delay(1000);
    }
  }
  Serial.print("unresistered ID,ID=");
  Serial.println(ID);
  i = 0; // Clear sampling times
  /*Fingerprint sampling 3 times*/
  while (i < COLLECT_NUMBER) {
    /*Set fingerprint LED ring mode, color, and number of blinks
      Can be set as follows:
      Parameter 1:<LEDMode>
      eBreathing   eFastBlink   eKeepsOn    eNormalClose
      eFadeIn      eFadeOut     eSlowBlink
      Parameter 2:<LEDColor>
      eLEDGreen  eLEDRed      eLEDYellow   eLEDBlue
      eLEDCyan   eLEDMagenta  eLEDWhite
      Parameter 3:<Number of blinks> 0 represents blinking all the time
      This parameter will only be valid in mode eBreathing, eFastBlink,
      eSlowBlink
     */
    fingerprint.ctrlLED(/*LEDMode = */ fingerprint.eBreathing,
                        /*LEDColor = */ fingerprint.eLEDBlue,
                        /*blinkCount = */ 0);
    Serial.print("The fingerprint sampling of the");
    Serial.print(i + 1);
    Serial.println("(th) is being taken");
    Serial.println("Please press down your finger");
    /*Capture fingerprint image, 10s idle timeout, if timeout=0,Disable  the
      collection timeout function IF succeeded, return 0, otherwise, return
      ERR_ID809
     */
    if ((fingerprint.collectionFingerprint(/*timeout = */ 10)) != ERR_ID809) {
      /*Set fingerprint LED ring to quick blink in yellow 3 times */
      fingerprint.ctrlLED(/*LEDMode = */ fingerprint.eFastBlink,
                          /*LEDColor = */ fingerprint.eLEDYellow,
                          /*blinkCount = */ 3);
      Serial.println("Sampling succeeds");
      i++; // Sampling times +1
    } else {
      Serial.println("Sampling failed");
      /*Get error code information*/
      // desc = fingerprint.getErrorDescription();
      // Serial.println(desc);
    }
    Serial.println("Please release your finger");
    /*Wait for finger to release
      Return 1 when finger is detected, otherwise return 0
     */
    while (fingerprint.detectFinger())
      ;
  }

  if (ID < 1) {
    Serial.println("Fingerprint registration failed");
    return;
  }

  /*Save fingerprint in an unregistered ID */
  if (fingerprint.storeFingerprint(/*Empty ID = */ ID) != ERR_ID809) {
    Serial.print("Saving succeed，ID=");
    Serial.println(ID);
    Serial.println("-----------------------------");
    /*Set fingerprint LED ring to always ON in green */
    fingerprint.ctrlLED(/*LEDMode = */ fingerprint.eKeepsOn,
                        /*LEDColor = */ fingerprint.eLEDGreen,
                        /*blinkCount = */ 0);
    delay(1000);
    /*Turn off fingerprint LED ring */
    fingerprint.ctrlLED(/*LEDMode = */ fingerprint.eNormalClose,
                        /*LEDColor = */ fingerprint.eLEDBlue,
                        /*blinkCount = */ 0);
    delay(1000);

    fingerprint.getTemplate(ID, temp);
    compute_hash();
    saveHash();
    readHash();

  } else {
    Serial.println("Saving failed");
    /*Get error code information*/
    // desc = fingerprint.getErrorDescription();
    // Serial.println(desc);
  }
}

/*-----------Arduino loop function--------- */

void loop() {
  blink_led();
  regFingerprint();
}
